import math
import os
from typing import Optional

import numpy as np
import timm
import torch
import torch.nn as nn
from torch import optim
from torch.profiler import profile
from torchmetrics.classification import BinaryAccuracy
from tqdm import tqdm
from transformers import get_cosine_schedule_with_warmup, get_wsd_schedule

from vidarr.dali import dali_train_loader, dali_val_loader
from vidarr.utils import freeze_model, print_rank_0, timed

torch.set_float32_matmul_precision("high")
torch.backends.cudnn.benchmark = True
torch.backends.cuda.matmul.allow_tf32 = True


def save_checkpoint(model, save_dir: str, filename: str = "final_model.pt"):
    if not os.path.exists(save_dir):
        os.makedirs(save_dir, exist_ok=True)

    save_path = os.path.join(save_dir, filename)
    state_dict = model.state_dict()
    print_rank_0(f"Saving final checkpoint to {save_path}...")
    torch.save(state_dict, save_path)


def load_model(
    model_name: str,
    num_classes: int = 1,
    drop_rate: float = 0.1,
    drop_path_rate: Optional[float] = 0.05,
    train_classification_head: bool = False,
    device: str = "cuda",
):
    device = torch.device(device=device)
    model = timm.create_model(
        model_name=model_name,
        pretrained=True,
        num_classes=num_classes,
        drop_rate=drop_rate,
        drop_path_rate=drop_path_rate,
    ).to(device)
    print_rank_0("=" * 80)
    print_rank_0(model)
    print_rank_0("=" * 80)

    if train_classification_head:
        model = freeze_model(model=model)
        print_rank_0("=" * 80)
    return model


def load_optimizer(
    model,
    lr: float,
    fused: bool = True,
    foreach: Optional[bool] = None,
    capturable: bool = False,
):
    optimizer = optim.AdamW(
        filter(lambda p: p.requires_grad, model.parameters()),
        lr=lr,
        fused=fused,
        foreach=foreach,
        capturable=capturable,
    )
    return optimizer


@torch.compile
def load_criterion(criterion_type: str):
    if criterion_type == "bcewithlogits":
        criterion = nn.BCEWithLogitsLoss()
    else:
        raise NotImplementedError()
    return criterion


def load_metric(metric_type: str, device="cuda"):
    if metric_type == "binary":
        metric = BinaryAccuracy()
    else:
        raise NotImplementedError()
    return metric.to(device=device)


def load_lr_scheduler(
    optimizer,
    scheduler_type: str,
    total_training_steps: int,
    warmup_steps: float,
    decay_steps: Optional[float] = None,
):
    num_warmup_steps = int(total_training_steps * warmup_steps)

    if scheduler_type == "wsd":
        num_decay_steps = int(total_training_steps * decay_steps)
        num_train_steps = total_training_steps - num_warmup_steps - num_decay_steps
        lr_scheduler = get_wsd_schedule(
            optimizer=optimizer,
            num_warmup_steps=num_warmup_steps,
            num_training_steps=num_train_steps,
            num_decay_steps=num_decay_steps,
        )
    elif scheduler_type == "cosine":
        num_train_steps = total_training_steps - num_warmup_steps
        lr_scheduler = get_cosine_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=num_warmup_steps,
            num_training_steps=num_train_steps,
        )
    else:
        raise NotImplementedError()
    return lr_scheduler


def _step(
    model,
    optimizer,
    criterion,
    lr_scheduler,
    inputs,
    labels,
    scaler,
    metric,
    is_train: bool,
):
    if is_train:
        optimizer.zero_grad(set_to_none=True)

    with torch.amp.autocast(
        device_type="cuda",
        dtype=torch.float16 if scaler is not None else torch.bfloat16,
    ):
        pred = model(inputs)
        loss = criterion(pred, labels)

    if metric:
        metric.update(pred, labels)

    if is_train:
        if scaler:
            scaler.scale(loss).backward()
            scaler.step(optimizer)
            old_scaler = scaler.get_scale()
            scaler.update()
            new_scaler = scaler.get_scale()
            if new_scaler >= old_scaler:
                lr_scheduler.step()
        else:
            loss.backward()
            optimizer.step()
            lr_scheduler.step()

    return loss


def _epoch(
    model,
    dataloader,
    optimizer,
    criterion,
    lr_scheduler,
    scaler,
    metric,
    prof,
    epoch,
    is_train: bool,
):
    model.train() if is_train else model.eval()
    metric.reset()
    timed_steps = []
    epoch_loss = 0
    desc = "Training" if is_train else "Evaluating"
    pbar = tqdm(total=len(dataloader), desc=desc)
    with torch.set_grad_enabled(is_train):
        for step, batch_data in enumerate(dataloader):
            torch.compiler.cudagraph_mark_step_begin()
            inputs = batch_data[0]["data"]  # Shape: [B, C, H, W]
            labels = batch_data[0]["label"].float()
            loss, times = timed(
                lambda: _step(
                    model,
                    optimizer,
                    criterion,
                    lr_scheduler,
                    inputs,
                    labels,
                    scaler,
                    metric,
                    is_train,
                )
            )
            prof.step()
            timed_steps.append(times)
            epoch_loss += loss.item()
            stats = {
                "epoch": (epoch + 1),
                "accuracy": metric.compute().item(),
                "avg_loss": epoch_loss / (step + 1),
                "median step time": np.median(timed_steps),
            }
            if is_train:
                stats["lr"] = lr_scheduler.get_last_lr()[0]
            pbar.set_postfix(stats)
            pbar.update()

    dataloader.reset()
    pbar.close()
    return epoch_loss / (step + 1)


def run_training(
    model,
    train_dataloader,
    val_dataloader,
    optimizer,
    criterion,
    lr_scheduler,
    scaler,
    metric,
    num_epochs,
    profiler_dir,
    checkpoint_dir,
):
    with profile(
        schedule=torch.profiler.schedule(wait=1, warmup=1, active=3, repeat=1),
        on_trace_ready=torch.profiler.tensorboard_trace_handler(dir_name=profiler_dir),
        record_shapes=True,
        profile_memory=True,
        with_stack=True,
    ) as prof:
        for epoch in tqdm(range(num_epochs), desc="Epochs"):
            train_loss = _epoch(
                model=model,
                dataloader=train_dataloader,
                optimizer=optimizer,
                criterion=criterion,
                lr_scheduler=lr_scheduler,
                scaler=scaler,
                metric=metric,
                prof=prof,
                epoch=epoch,
                is_train=True,
            )
            val_loss = _epoch(
                model=model,
                dataloader=val_dataloader,
                optimizer=None,
                criterion=criterion,
                lr_scheduler=None,
                scaler=scaler,
                metric=metric,
                prof=prof,
                epoch=epoch,
                is_train=False,
            )

    save_checkpoint(model=model, save_dir=checkpoint_dir)


def train(
    model_name: str,
    train_dir: str,
    val_dir: str,
    num_epochs: int,
    batch_size: int,
    learning_rate: float,
    num_classes: int = 1,
    drop_rate: Optional[float] = 0.1,
    drop_path_rate: Optional[float] = None,
    scheduler_type: str = "cosine",
    warmup_steps: float = 0.10,
    decay_steps: Optional[float] = 0.10,
    num_threads: int = 4,
    image_size: int = 224,
    image_crop: int = 224,
    use_scaler: bool = False,
    use_compile: bool = False,
    train_classification_head: bool = False,
    metric_type: str = "binary",
    criterion_type: str = "bcewithlogits",
    profiler_dir: str = "./log",
    checkpoint_dir: str = "./checkpoints",
):
    train_dataloader = dali_train_loader(
        images_dir=train_dir,
        batch_size=batch_size,
        num_threads=num_threads,
        image_size=image_size,
        image_crop=image_crop,
    )

    val_dataloader = dali_val_loader(
        images_dir=val_dir,
        batch_size=batch_size,
        num_threads=num_threads,
        image_size=image_size,
        image_crop=image_crop,
    )

    model = load_model(
        model_name=model_name,
        num_classes=num_classes,
        drop_path_rate=drop_path_rate,
        drop_rate=drop_rate,
        train_classification_head=train_classification_head,
    )
    if use_compile:
        model = torch.compile(model)

    metric = load_metric(metric_type=metric_type)
    optimizer = load_optimizer(model=model, lr=learning_rate)
    criterion = load_criterion(criterion_type=criterion_type)
    lr_scheduler = load_lr_scheduler(
        optimizer=optimizer,
        scheduler_type=scheduler_type,
        total_training_steps=num_epochs * len(train_dataloader),
        warmup_steps=warmup_steps,
        decay_steps=decay_steps,
    )

    if use_scaler:
        scaler = torch.amp.GradScaler()
    else:
        scaler = None

    run_training(
        model=model,
        train_dataloader=train_dataloader,
        val_dataloader=val_dataloader,
        optimizer=optimizer,
        criterion=criterion,
        lr_scheduler=lr_scheduler,
        scaler=scaler,
        metric=metric,
        num_epochs=num_epochs,
        profiler_dir=profiler_dir,
        checkpoint_dir=checkpoint_dir,
    )



from typing import Dict

from hta.trace_analysis import TraceAnalysis


def analyze_run(
    breakdown: str,
    trace_dir: str,
    trace_files: Dict[int, str] | None = None,
):
    analyzer = TraceAnalysis(trace_dir=trace_dir, trace_files=trace_files)

    if breakdown == "temporal":
        return analyzer.get_temporal_breakdown()
    elif breakdown == "idle":
        return analyzer.get_idle_time_breakdown()
    else:
        raise NotImplementedError()



import nvidia.dali.fn as fn
import nvidia.dali.types as types
from nvidia.dali.auto_aug import auto_augment
from nvidia.dali.pipeline import pipeline_def
from nvidia.dali.plugin.pytorch import DALIClassificationIterator


def apply_training_augmentations(images, image_size: int, image_crop: int):
    images = fn.decoders.image_random_crop(
        images, device="mixed", output_type=types.RGB
    )

    images = fn.resize(images, size=image_size)

    images = images.gpu()

    rng = fn.random.coin_flip(probability=0.5)
    images = fn.flip(images, horizontal=rng)

    images = auto_augment.auto_augment_image_net(images, shape=[image_size, image_size])

    images = fn.crop_mirror_normalize(
        images,
        dtype=types.FLOAT,
        crop=(image_crop, image_crop),
        mean=[0.485 * 255, 0.456 * 255, 0.406 * 255],
        std=[0.229 * 255, 0.224 * 255, 0.225 * 255],
    )
    return images


def apply_validation_augmentations(images, image_size: int, image_crop: int):
    images = fn.decoders.image(images, device="mixed", output_type=types.RGB)

    images = fn.resize(images, size=image_size)

    images = images.gpu()

    images = fn.crop_mirror_normalize(
        images,
        dtype=types.FLOAT,
        crop=(image_crop, image_crop),
        mean=[0.485 * 255, 0.456 * 255, 0.406 * 255],
        std=[0.229 * 255, 0.224 * 255, 0.225 * 255],
    )
    return images


@pipeline_def(enable_conditionals=True)
def dali_training_pipeline(images_dir: str, image_size: int, image_crop: int):
    images, labels = fn.readers.file(
        file_root=images_dir, random_shuffle=True, pad_last_batch=True, name="Reader"
    )
    images = apply_training_augmentations(
        images=images, image_size=image_size, image_crop=image_crop
    )
    return images, labels.gpu()


@pipeline_def
def dali_validation_pipeline(images_dir: str, image_size: int, image_crop: int):
    images, labels = fn.readers.file(
        file_root=images_dir, random_shuffle=False, pad_last_batch=True, name="Reader"
    )
    images = apply_validation_augmentations(
        images=images, image_size=image_size, image_crop=image_crop
    )
    return images, labels.gpu()


def dali_train_loader(
    images_dir: str,
    batch_size: int,
    num_threads: int = 4,
    device_id: int = 0,
    image_size: int = 224,
    image_crop: int = 224,
):
    pipeline_kwargs = {
        "batch_size": batch_size,
        "num_threads": num_threads,
        "device_id": device_id,
    }
    pipe = dali_training_pipeline(
        images_dir=images_dir,
        image_size=image_size,
        image_crop=image_crop,
        **pipeline_kwargs,
    )
    train_loader = DALIClassificationIterator(
        pipe,
        reader_name="Reader",
    )
    return train_loader


def dali_val_loader(
    images_dir: str,
    batch_size: int,
    num_threads: int = 4,
    device_id: int = 0,
    image_size: int = 224,
    image_crop: int = 224,
):
    pipeline_kwargs = {
        "batch_size": batch_size,
        "num_threads": num_threads,
        "device_id": device_id,
    }
    pipe = dali_validation_pipeline(
        images_dir=images_dir,
        image_size=image_size,
        image_crop=image_crop,
        **pipeline_kwargs,
    )
    val_loader = DALIClassificationIterator(
        pipe,
        reader_name="Reader",
    )
    return val_loader



import torch

model = torch.nn.Sequential(
    *[torch.nn.Linear(1024, 1024, False, device="cuda") for _ in range(10)]
)
input = torch.rand(1024, device="cuda")
output = model(input)
output.sum().backward()

# exit cleanly if we are on a device that doesn't support torch.compile
if torch.cuda.get_device_capability() < (7, 0):
    print("Exiting because torch.compile is not supported on this device.")
    import sys

    sys.exit(0)


optimizer = torch.optim.Adam(model.parameters(), lr=0.01)


def eager_step():
    optimizer.step()


def compiled_step():
    optimizer.step()


compiled_step = torch.compile(compiled_step, fullgraph=False)

# Let's define a helpful benchmarking function:
import torch.utils.benchmark as benchmark


def benchmark_torch_function_in_microseconds(f, *args, **kwargs):
    t0 = benchmark.Timer(
        stmt="f(*args, **kwargs)", globals={"args": args, "kwargs": kwargs, "f": f}
    )
    return t0.blocked_autorange().mean * 1e6


# Warmup runs to compile the function
for _ in range(5):
    compiled_step()

eager_runtime = benchmark_torch_function_in_microseconds(eager_step)
compiled_runtime = benchmark_torch_function_in_microseconds(compiled_step)

assert eager_runtime > compiled_runtime

print(f"eager runtime: {eager_runtime}us")
print(f"compiled runtime: {compiled_runtime}us")



import torch
import torch.nn as nn
import wandb


def freeze_model(model):
    for p in model.parameters():
        p.requires_grad = False

    linear_layers = [m for m in model.modules() if isinstance(m, nn.Linear)]
    if not linear_layers:
        raise ValueError("No Linear layers found in model.")

    last_linear = linear_layers[-1]

    for p in last_linear.parameters():
        p.requires_grad = True

    trainable = [n for n, p in model.named_parameters() if p.requires_grad]
    print_rank_0("Trainable parameters:", trainable)
    return model


def timed(fn):
    start_event = torch.cuda.Event(enable_timing=True)
    end_event = torch.cuda.Event(enable_timing=True)
    start_event.record()
    result = fn()
    end_event.record()
    torch.cuda.synchronize()
    training_time_ms = start_event.elapsed_time(end_event)
    training_time_s = training_time_ms / 1000.0
    return result, training_time_s


def model_num_params(model):
    non_embedding_params = sum(
        param.numel() for name, param in model.named_parameters() if "embed" not in name
    )
    return non_embedding_params


def initialize_writer(
    entity: str,
    project: str,
    global_batch_size: int,
    learning_rate: float,
    num_training_epochs: int,
    gradient_accumulation_steps: int,
    model_name: str,
    dataset_name: str,
):
    writer = wandb.init(
        entity=entity,
        project=project,
        name=f"BS: {global_batch_size} LR: {learning_rate}",
        config={
            "learning_rate": learning_rate,
            "model": model_name,
            "dataset": dataset_name,
            "epochs": num_training_epochs,
            "gradient accumulation steps": gradient_accumulation_steps,
        },
    )
    return writer


def print_rank_0(message, rank=None):
    """If distributed is initialized or rank is specified, print only on rank 0."""
    if rank is not None:
        if rank == 0:
            print(message, flush=True)
    elif torch.distributed.is_initialized():
        if torch.distributed.get_rank() == 0:
            print(message, flush=True)
    else:
        print(message, flush=True)



import albumentations as A
import lightly_train

if __name__ == "__main__":
    lightly_train.train(
        out="out/my_experiment",
        data="my_data_dir",
        model="timm/efficientvit_m0.r224_in1k",
        method="distillation",
        epochs=100,
        batch_size=128,
    )
